from __future__ import division
import numpy as np
from .linalg import dot_inplace_right


def eigenvalue_decomposition(S, eps=10**-10):
    r"""
    Eigenvalue decomposition of a given covariance (or scatter) matrix.

    Parameters
    ----------
    S : ``(N, N)`` `ndarray`
        Covariance/Scatter matrix
    eps : `float`, optional
        Small value to be used for the tolerance limit computation. The final
        limit is computed as ::

            limit = np.max(np.abs(eigenvalues)) * eps

    Returns
    -------
    pos_eigenvectors : ``(N, p)`` `ndarray`
        The matrix with the eigenvectors corresponding to positive eigenvalues.
    pos_eigenvalues : ``(p,)`` `ndarray`
        The array of positive eigenvalues.
    """
    # compute eigenvalue decomposition
    eigenvalues, eigenvectors = np.linalg.eigh(S)
    # sort eigenvalues from largest to smallest
    index = np.argsort(eigenvalues)[::-1]
    eigenvalues = eigenvalues[index]
    eigenvectors = eigenvectors[:, index]

    # set tolerance limit
    limit = np.max(np.abs(eigenvalues)) * eps

    # select positive eigenvalues
    pos_index = eigenvalues > 0.0
    pos_eigenvalues = eigenvalues[pos_index]
    pos_eigenvectors = eigenvectors[:, pos_index]
    # check they are within the expected tolerance
    index = pos_eigenvalues > limit
    pos_eigenvalues = pos_eigenvalues[index]
    pos_eigenvectors = pos_eigenvectors[:, index]

    return pos_eigenvectors, pos_eigenvalues


def principal_component_decomposition(X, whiten=False, centre=True,
                                      bias=False, inplace=False):
    r"""
    Apply Principal Component Analysis (PCA) on the data matrix `X`. In the case
    where the data matrix is very large, it is advisable to set
    ``inplace = True``. However, note this destructively edits the data matrix
    by subtracting the mean inplace.

    Parameters
    ----------
    X : ``(n_samples, n_features)`` `ndarray`
        Training data.
    whiten : `bool`, optional
        Normalise the eigenvectors to have unit magnitude.
    centre : `bool`, optional
        Whether to centre the data matrix. If ``False``, zero will be
        subtracted.
    bias : `bool`, optional
        Whether to use a biased estimate of the number of samples. If ``False``,
        subtracts ``1`` from the number of samples.
    inplace : `bool`, optional
        Whether to do the mean subtracting inplace or not. This is crucial if
        the data matrix is greater than half the available memory size.

    Returns
    -------
    eigenvectors : ``(n_components, n_features)`` `ndarray`
        The eigenvectors of the data matrix.
    eigenvalues : ``(n_components,)`` `ndarray`
        The positive eigenvalues from the data matrix.
    mean_vector : ``(n_components,)`` `ndarray`
        The mean that was subtracted from the dataset.
    """
    n_samples, n_features = X.shape

    if bias:
        N = n_samples
    else:
        N = n_samples - 1.0

    if centre:
        # centre data
        mean_vector = np.mean(X, axis=0)
    else:
        mean_vector = np.zeros(n_features)

    # This is required if the data matrix is very large!
    if inplace:
        X -= mean_vector
    else:
        X = X - mean_vector

    if n_features < n_samples:
        # compute covariance matrix
        # S:  n_features  x  n_features
        S = np.dot(X.T, X) / N
        # S should be perfectly symmetrical, but numerical error can creep
        # in. Enforce symmetry here to avoid creating complex
        # eigenvectors from eigendecomposition
        S = (S + S.T) / 2.0

        # perform eigenvalue decomposition
        # eigenvectors:  n_features x  n_features
        # eigenvalues:   n_features
        eigenvectors, eigenvalues = eigenvalue_decomposition(S)

        if whiten:
            # whiten eigenvectors
            eigenvectors *= np.sqrt(1.0 / eigenvalues)

        # transpose eigenvectors
        # eigenvectors:  n_samples  x  n_features
        eigenvectors = eigenvectors.T

    else:
        # n_features > n_samples
        # compute covariance matrix
        # S:  n_samples  x  n_samples
        S = np.dot(X, X.T) / N
        # S should be perfectly symmetrical, but numerical error can creep
        # in. Enforce symmetry here to avoid creating complex
        # eigenvectors from eigendecomposition
        S = (S + S.T) / 2.0

        # perform eigenvalue decomposition
        # eigenvectors:  n_samples  x  n_samples
        # eigenvalues:   n_samples
        eigenvectors_s, eigenvalues = eigenvalue_decomposition(S)

        # compute final eigenvectors
        # eigenvectors:  n_samples  x  n_features
        if whiten:
            w = (N * eigenvalues) ** -1.0
        else:
            w = np.sqrt(1.0 / (N * eigenvalues))

        dot = dot_inplace_right if inplace else np.dot
        eigenvectors = dot(eigenvectors_s.T, X)

        # whiten, and we are done.
        eigenvectors *= w[:, None]

    return eigenvectors, eigenvalues, mean_vector



# distutils: language = c
# distutils: sources = ./menpo/transform/piecewiseaffine/fastpwa/pwa.c

import numpy as np
cimport numpy as cnp

cdef extern from "./fastpwa/pwa.h":
    ctypedef struct TriangleCollection:
        pass

    ctypedef struct AlphaBetaIndex:
        pass

    TriangleCollection initTriangleCollection(double *vertices,
                                              unsigned int *trilist,
                                              unsigned int n_triangles)

    void arrayCachedAlphaBetaIndexForPoints(AlphaBetaIndex **hashMap,
                                      TriangleCollection *tris,
                                      double *points,
                                      unsigned int n_points, int *indexes,
                                      double *alphas, double *betas)
    void arrayAlphaBetaIndexForPoints(TriangleCollection *tris,
                                      double *points,
                                      unsigned int n_points, int *indexes,
                                      double *alphas, double *betas)
    void clearCacheAndDelete(AlphaBetaIndex **hashMap)
    void deleteTriangleCollection(TriangleCollection *tris)

cdef class CLookupPWA:
    cdef TriangleCollection tris
    cdef AlphaBetaIndex *hashMap
    cdef unsigned n_tris
    cdef object points
    cdef object trilist

    def __cinit__(self,
                  double[:, ::1] points not None,
                  unsigned[:, ::1] trilist not None):
        hashMap = NULL
        if points.shape[1] != 2:
            raise Exception
        self.n_tris = trilist.shape[0]
        self.points = points
        self.trilist = trilist
        self.tris =  initTriangleCollection(&points[0,0], &trilist[0,0],
                                            trilist.shape[0])

    def _init_source_triangles(self,
                  double[:, ::1] points not None,
                  unsigned[:, ::1] trilist not None):
        hashMap = NULL
        if points.shape[1] != 2:
            raise Exception
        elif points.shape[0] != self.n_tris:
            raise Exception
        self.tris =  initTriangleCollection(&points[0,0], &trilist[0,0],
                                            self.n_tris)

    def _init_target_triangles(
            self, double[:, ::1] points not None,
            unsigned[:, ::1] trilist not None):
        self.target_tris = initTriangleCollection(
            &points[0, 0], &trilist[0, 0], trilist.shape[0])

    def __dealloc__(self):
        deleteTriangleCollection(&self.tris)
        clearCacheAndDelete(&self.hashMap)

    def __reduce__(self):
        r"""
        Implement the reduction protocol so this object is copyable/picklable
        """
        return self.__class__, (np.asarray(self.points),
                                np.asarray(self.trilist))

    def index_alpha_beta(self, double[:, ::1] points not None):
        # create three c numpy arrays for storing our output into
        cdef cnp.ndarray[double, ndim=1, mode='c'] alphas = \
            np.zeros(points.shape[0], dtype=np.float64)
        cdef cnp.ndarray[double, ndim=1, mode='c'] betas = \
            np.zeros(points.shape[0], dtype=np.float64)
        cdef cnp.ndarray[int, ndim=1, mode='c'] indexes = \
            np.zeros(points.shape[0], dtype=np.int32)
        # fill the arrays with the C results
        arrayCachedAlphaBetaIndexForPoints(&self.hashMap, &self.tris,
                                          &points[0,0],
                                     points.shape[0], &indexes[0],
                                     &alphas[0], &betas[0])
        return indexes, alphas, betas



# distutils: language = c++
import numpy as np
cimport numpy as np
cimport cython


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void calc_augmented_centres(const double[:, :] centres,
                                 const np.int64_t[:, :] sample_offsets,
                                 np.int64_t[:, :] augmented_centres):
    r"""
    For each centre that was given (centre of a patch), generate another
    patch that is an offset away from that centre. This is useful for generating
    a dense set of patch based features around a given point.

    Parameters (Inputs)
    -------------------
    centres : double[:, :] (n_points, 2)
        The centres of each patch.
    sample_offsets : np.int64_t[:, :] (n_sample_offsets, 2)
        The 2D offsets to sample extra patches around

    Parameters (Outputs)
    --------------------
    augmented_centres : np.int64_t[:, :] (n_points * n_sample_offsets, 2)
        The output buffer
    """
    cdef:
        np.int64_t total_index = 0, i = 0, j = 0

    for i in range(centres.shape[0]):
        for j in range(sample_offsets.shape[0]):
            augmented_centres[total_index, 0] = <np.int64_t> (centres[i, 0] + sample_offsets[j, 0])
            augmented_centres[total_index, 1] = <np.int64_t> (centres[i, 1] + sample_offsets[j, 1])
            total_index += 1

@cython.boundscheck(False)
@cython.wraparound(False)
cdef void calc_slices(const np.int64_t[:, :] centres,
                      np.int64_t image_shape0,
                      np.int64_t image_shape1,
                      np.int64_t patch_shape0,
                      np.int64_t patch_shape1,
                      np.int64_t[:, :] ext_s_min,
                      np.int64_t[:, :] ext_s_max,
                      np.int64_t[:, :] ins_s_min,
                      np.int64_t[:, :] ins_s_max):
    r"""
    For each centre that was given (centre of a patch), generate a slice in to
    the original image that represents the bounds of the patch. This method
    also handles ensuring that all patches have a correct set of bounds. Patches
    that would lie outside that bounds of the image are truncated back inside
    the image.

    Parameters (Inputs)
    -------------------
    centres : double[:, :] (n_points, 2)
        The centres of each patch.
    image_shape0 : np.int64_t
        The size of the first dimension of the image (height)
    image_shape1 : np.int64_t
        The size of the second dimension of the image (width)
    patch_shape0 : np.int64_t
        The size of the first dimension of the patch (height)
    patch_shape1 : np.int64_t
        The size of the second dimension of the patch (width)

    Parameters (Outputs)
    --------------------
    ext_s_min : np.int64_t[:, :] (n_centres, 2)
        The extraction slice minimum indices. This is in the image domain, one
        for each slice.
    ext_s_max : np.int64_t[:, :] (n_centres, 2)
        The extraction slice maximum indices. This is in the image domain, one
        for each slice.
    ins_s_min : np.int64_t[:, :] (n_centres, 2)
        The insertion slice minimum indices. This is in the patch domain, one
        for each slice.
    ins_s_max : np.int64_t[:, :] (n_centres, 2)
        The insertion slice maximum indices. This is in the patch domain, one
        for each slice.
    """
    cdef:
        np.int64_t c_min_new0 = 0, c_min_new1 = 0, c_max_new0 = 0, c_max_new1 = 0, i = 0
        np.int64_t half_patch_shape0 = patch_shape0 / 2
        np.int64_t half_patch_shape1 = patch_shape1 / 2
        np.int64_t add_to_patch0 = patch_shape0 % 2
        np.int64_t add_to_patch1 = patch_shape1 % 2

    for i in range(centres.shape[0]):
        c_min_new0 = centres[i, 0] - half_patch_shape0
        c_min_new1 = centres[i, 1] - half_patch_shape1
        c_max_new0 = centres[i, 0] + half_patch_shape0 + add_to_patch0
        c_max_new1 = centres[i, 1] + half_patch_shape1 + add_to_patch1

        ext_s_min[i, 0] = c_min_new0
        ext_s_min[i, 1] = c_min_new1
        ext_s_max[i, 0] = c_max_new0
        ext_s_max[i, 1] = c_max_new1

        if ext_s_min[i, 0] < 0:
            ext_s_min[i, 0] = 0
        if ext_s_min[i, 1] < 0:
            ext_s_min[i, 1] = 0
        if ext_s_min[i, 0] > image_shape0:
            ext_s_min[i, 0] = image_shape0 - 1
        if ext_s_min[i, 1] > image_shape1:
            ext_s_min[i, 1] = image_shape1 - 1

        if ext_s_max[i, 0] < 0:
            ext_s_max[i, 0] = 0
        if ext_s_max[i, 1] < 0:
            ext_s_max[i, 1] = 0
        if ext_s_max[i, 0] > image_shape0:
            ext_s_max[i, 0] = image_shape0 - 1
        if ext_s_max[i, 1] > image_shape1:
            ext_s_max[i, 1] = image_shape1 - 1

        ins_s_min[i, 0] = ext_s_min[i, 0] - c_min_new0
        ins_s_min[i, 1] = ext_s_min[i, 1] - c_min_new1

        ins_s_max[i, 0] = ext_s_max[i, 0] - c_max_new0 + patch_shape0
        if ins_s_max[i, 0] < 0:
            ins_s_max[i, 0] = 0
        ins_s_max[i, 1] = ext_s_max[i, 1] - c_max_new1 + patch_shape1
        if ins_s_max[i, 1] < 0:
            ins_s_max[i, 1] = 0


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void slice_image(const double[:, :, :] image,
                      const np.int64_t n_centres,
                      const np.int64_t n_sample_offsets,
                      const np.int64_t[:, :] ext_s_min,
                      const np.int64_t[:, :] ext_s_max,
                      const np.int64_t[:, :] ins_s_min,
                      const np.int64_t[:, :] ins_s_max,
                      double[:, :, :, :] patches):
    r"""
    Extract all the patches from the image. The patch extents have already been
    calculated and so this function simply slices appropriately in to the image
    to extract the pixels.

    Parameters (Inputs)
    -------------------
    image : double[:, :, :] (height, width, n_channels)
        The image to extract patches from.
    n_centres : np.int64_t
        The number of centres given.
    n_sample_offsets : np.int64_t
        The number of sample offsets given.
    ext_s_min : np.int64_t[:, :] (n_centres, 2)
        The extraction slice minimum indices. This is in the image domain, one
        for each slice.
    ext_s_max : np.int64_t[:, :] (n_centres, 2)
        The extraction slice maximum indices. This is in the image domain, one
        for each slice.
    ins_s_min : np.int64_t[:, :] (n_centres, 2)
        The insertion slice minimum indices. This is in the patch domain, one
        for each slice.
    ins_s_max : np.int64_t[:, :] (n_centres, 2)
        The insertion slice maximum indices. This is in the patch domain, one
        for each slice.

    Parameters (Outputs)
    --------------------
    patches : np.int64_t[:, :, :, :] (n_centres * n_sample_offsets, height, width, channels)
        The set of patches that have been extracted.
    """
    cdef np.int64_t i = 0
    for i in range(n_centres * n_sample_offsets):
            patches[i,
                    ins_s_min[i, 0]:ins_s_max[i, 0],
                    ins_s_min[i, 1]:ins_s_max[i, 1]] = \
            image[ext_s_min[i, 0]:ext_s_max[i, 0],
                  ext_s_min[i, 1]:ext_s_max[i, 1]]


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef extract_patches_cython(double[:, :, :] image, double[:, :] centres,
                             np.int64_t[:] patch_size, np.int64_t[:, :] sample_offsets):
    r"""
    Extract a set of patches from an image. Given a set of patch centres and
    a patch size, patches are extracted from within the image centreed
    on the given coordinates. Sample offsets denote a set of offsets to extract
    from within a patch. This is very useful if you want to extract a dense
    set of features around a set of landmarks and simply sample the same grid
    of patches around the landmarks.

    Parameters
    ----------
    image : double[:, :, :] (height, width, n_channels)
        The image to extract patches from.
    n_centres : np.int64_t
        The number of centres given.
    n_sample_offsets : np.int64_t
        The number of sample offsets given.
    ext_s_min : np.int64_t[:, :] (n_centres, 2)
        The extraction slice minimum indices. This is in the image domain, one
        for each slice.
    ext_s_max : np.int64_t[:, :] (n_centres, 2)
        The extraction slice maximum indices. This is in the image domain, one
        for each slice.
    ins_s_min : np.int64_t[:, :] (n_centres, 2)
        The insertion slice minimum indices. This is in the patch domain, one
        for each slice.
    ins_s_max : np.int64_t[:, :] (n_centres, 2)
        The insertion slice maximum indices. This is in the patch domain, one
        for each slice.

    Returns
    -------
    patches : double[:, :, :, :] (n_centres * n_sample_offsets, height, width, n_channels)
        Returns a matrix containing the set of patches. The last 3 dimensions
        are the patches and the first dimension is each centre. If multiple
        sample offsets were provided, then they are all concatenated together
        to return (n_centres * n_sample_offsets) patches.
    """
    cdef:
        np.int64_t n_centres = centres.shape[0]
        np.int64_t n_sample_offsets = sample_offsets.shape[0]
        np.int64_t n_augmented_centres = n_centres * n_sample_offsets

        np.int64_t[:, :] augmented_centres = np.empty([n_augmented_centres, 2],
                                                dtype=np.int64)

        np.int64_t patch_shape0 = patch_size[0]
        np.int64_t patch_shape1 = patch_size[1]
        np.int64_t image_shape0 = image.shape[0]
        np.int64_t image_shape1 = image.shape[1]
        np.int64_t n_channels = image.shape[2]

        # This could be faster with malloc
        np.int64_t[:,:] ext_s_max = np.empty([n_augmented_centres, 2], dtype=np.int64)
        np.int64_t[:,:] ext_s_min = np.empty([n_augmented_centres, 2], dtype=np.int64)
        np.int64_t[:,:] ins_s_max = np.empty([n_augmented_centres, 2], dtype=np.int64)
        np.int64_t[:,:] ins_s_min = np.empty([n_augmented_centres, 2], dtype=np.int64)

        # It is important this array is zeros and not empty due to truncating
        # out of bounds patches.
        np.ndarray[double, ndim=4] patches = np.zeros(
            [n_centres * n_sample_offsets, patch_shape0, patch_shape1,
             n_channels])

    calc_augmented_centres(centres, sample_offsets, augmented_centres)
    calc_slices(augmented_centres,
                image_shape0,
                image_shape1,
                patch_shape0,
                patch_shape1,
                ext_s_min,
                ext_s_max,
                ins_s_min,
                ins_s_max)

    slice_image(image,
                n_centres,
                n_sample_offsets,
                ext_s_min,
                ext_s_max,
                ins_s_min,
                ins_s_max,
                patches)

    return patches





def dot_inplace_left(a, b, block_size=1000):
    r"""
    Inplace dot product for memory efficiency. It computes ``a * b = c``, where
    ``a`` will be replaced inplace with ``c``.

    Parameters
    ----------
    a : ``(n_big, k)`` `ndarray`
        First array to dot - assumed to be large. Will be damaged by this
        function call as it is used to store the output inplace.
    b : ``(k, n_small)`` `ndarray`, ``n_small <= k``
        The second array to dot - assumed to be small. ``n_small`` must be
        smaller than ``k`` so the result can be stored within the memory space
        of ``a``.
    block_size : `int`, optional
        The size of the block of ``a`` that will be dotted against ``b`` in
        each iteration. larger block sizes increase the time performance of the
        dot product at the cost of a higher memory overhead for the operation.

    Returns
    -------
    c : ``(n_big, n_small)`` `ndarray`
        The output of the operation. Exactly the same as a memory view onto
        ``a`` (``a[:, :n_small]``) as ``a`` is modified inplace to store the
        result.
    """
    (n_big, k_a), (k_b, n_small) = a.shape, b.shape
    if k_a != k_b:
        raise ValueError('Cannot dot {} * {}'.format(a.shape, b.shape))
    if n_small > k_a:
        raise ValueError('Cannot dot inplace left - '
                         'b.shape[1] ({}) > a.shape[1] '
                         '({})'.format(n_small, k_a))
    for i in range(0, n_big, block_size):
        j = i + block_size
        a[i:j, :n_small] = a[i:j].dot(b)
    return a[:, :n_small]


def dot_inplace_right(a, b, block_size=1000):
    r"""
    Inplace dot product for memory efficiency. It computes ``a * b = c`` where
    ``b`` will be replaced inplace with ``c``.

    Parameters
    ----------
    a : ``(n_small, k)`` `ndarray`, n_small <= k
        The first array to dot - assumed to be small. ``n_small`` must be
        smaller than ``k`` so the result can be stored within the memory space
        of ``b``.
    b : ``(k, n_big)`` `ndarray`
        Second array to dot - assumed to be large. Will be damaged by this
        function call as it is used to store the output inplace.
    block_size : `int`, optional
        The size of the block of ``b`` that ``a`` will be dotted against
        in each iteration. larger block sizes increase the time performance of
        the dot product at the cost of a higher memory overhead for the
        operation.

    Returns
    -------
    c : ``(n_small, n_big)`` `ndarray`
        The output of the operation. Exactly the same as a memory view onto
        ``b`` (``b[:n_small]``) as ``b`` is modified inplace to store the
        result.
    """
    (n_small, k_a), (k_b, n_big) = a.shape, b.shape
    if k_a != k_b:
        raise ValueError('Cannot dot {} * {}'.format(a.shape, b.shape))
    if n_small > k_b:
        raise ValueError('Cannot dot inplace right - '
                         'a.shape[1] ({}) > b.shape[0] '
                         '({})'.format(n_small, k_b))
    for i in range(0, n_big, block_size):
        j = i + block_size
        b[:n_small, i:j] = a.dot(b[:, i:j])
    return b[:n_small]



from .landmark import LJSONExporter, PTSExporter
from .image import PILExporter
from .pickle import pickle_export

landmark_types = {
    '.ljson': LJSONExporter,
    '.pts': PTSExporter
}

image_types = {
    '.bmp': PILExporter,
    '.dib': PILExporter,
    '.dcx': PILExporter,
    '.eps': PILExporter,
    '.ps': PILExporter,
    '.gif': PILExporter,
    '.im': PILExporter,
    '.jpg': PILExporter,
    '.jpe': PILExporter,
    '.jpeg': PILExporter,
    '.pcd': PILExporter,
    '.pcx': PILExporter,
    '.png': PILExporter,
    '.pbm': PILExporter,
    '.pgm': PILExporter,
    '.ppm': PILExporter,
    '.psd': PILExporter,
    '.tif': PILExporter,
    '.tiff': PILExporter,
    '.xbm': PILExporter,
    '.xpm': PILExporter
}

pickle_types = {
    '.pkl': pickle_export,
}
